#!/usr/bin/env python3

import logging as logging
from logging import handlers
import os


class Logger(logging.Logger):

    def __init__(self, name: str, logdir: str):
        super().__init__(name=name, level=logging.DEBUG)

        # create console handler and set level to info
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        # create file handler and set level to debug
        logfile = os.path.join(logdir, name + '.log')
        fh = handlers.TimedRotatingFileHandler(filename=logfile, when='D', interval=1, backupCount=6)
        fh.setLevel(logging.DEBUG)

        # create formatter
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        # add formatter to ch and fh
        ch.setFormatter(formatter)
        fh.setFormatter(formatter)

        # add ch and fh to logger
        self.addHandler(ch)
        self.addHandler(fh)



# -*- coding: utf-8 -*-
''' Handle Form Of Way (fow) rating

    .. moduleauthor:: David Marteau <david.marteau@mappy.com>

'''


from gv_vendor.pylr import fow

"""
Indicates an excellent match of the attributes of the line with the
attributes specified in the LRP
"""
EXCELLENT = 3

"""
Indicates a good match of the attributes of the line with the
attributes specified in the LRP
"""
GOOD = 2

"""
Indicates an average match of the attributes of the line with the
attributes specified in the LRP
"""
AVERAGE = 1

"""
Indicates a poor match of the attributes of the line with the
attributes specified in the LRP
"""
POOR = 0


_to_index = lambda a, b: a + 8*b

_FOWRating = 64 * [0]
_FOWRating[_to_index(fow.UNDEFINED, fow.UNDEFINED)] = AVERAGE
_FOWRating[_to_index(fow.UNDEFINED, fow.MOTORWAY)] = AVERAGE
_FOWRating[_to_index(fow.UNDEFINED, fow.MULTIPLE_CARRIAGEWAY)] = AVERAGE
_FOWRating[_to_index(fow.UNDEFINED, fow.SINGLE_CARRIAGEWAY)] = AVERAGE
_FOWRating[_to_index(fow.UNDEFINED, fow.ROUNDABOUT)] = AVERAGE
_FOWRating[_to_index(fow.UNDEFINED, fow.TRAFFICSQUARE)] = AVERAGE
_FOWRating[_to_index(fow.UNDEFINED, fow.SLIPROAD)] = AVERAGE
_FOWRating[_to_index(fow.UNDEFINED, fow.OTHER)] = AVERAGE

_FOWRating[_to_index(fow.MOTORWAY, fow.MOTORWAY)] = EXCELLENT
_FOWRating[_to_index(fow.MOTORWAY, fow.MULTIPLE_CARRIAGEWAY)] = GOOD
_FOWRating[_to_index(fow.MOTORWAY, fow.SINGLE_CARRIAGEWAY)] = POOR
_FOWRating[_to_index(fow.MOTORWAY, fow.ROUNDABOUT)] = POOR
_FOWRating[_to_index(fow.MOTORWAY, fow.TRAFFICSQUARE)] = POOR
_FOWRating[_to_index(fow.MOTORWAY, fow.SLIPROAD)] = POOR
_FOWRating[_to_index(fow.MOTORWAY, fow.OTHER)] = POOR

_FOWRating[_to_index(fow.MULTIPLE_CARRIAGEWAY, fow.MULTIPLE_CARRIAGEWAY)] = EXCELLENT
_FOWRating[_to_index(fow.MULTIPLE_CARRIAGEWAY, fow.SINGLE_CARRIAGEWAY)] = GOOD
_FOWRating[_to_index(fow.MULTIPLE_CARRIAGEWAY, fow.ROUNDABOUT)] = AVERAGE
_FOWRating[_to_index(fow.MULTIPLE_CARRIAGEWAY, fow.TRAFFICSQUARE)] = POOR
_FOWRating[_to_index(fow.MULTIPLE_CARRIAGEWAY, fow.SLIPROAD)] = POOR
_FOWRating[_to_index(fow.MULTIPLE_CARRIAGEWAY, fow.OTHER)] = POOR

_FOWRating[_to_index(fow.SINGLE_CARRIAGEWAY, fow.SINGLE_CARRIAGEWAY)] = EXCELLENT
_FOWRating[_to_index(fow.SINGLE_CARRIAGEWAY, fow.ROUNDABOUT)] = AVERAGE
_FOWRating[_to_index(fow.SINGLE_CARRIAGEWAY, fow.TRAFFICSQUARE)] = AVERAGE
_FOWRating[_to_index(fow.SINGLE_CARRIAGEWAY, fow.SLIPROAD)] = POOR
_FOWRating[_to_index(fow.SINGLE_CARRIAGEWAY, fow.OTHER)] = POOR

_FOWRating[_to_index(fow.ROUNDABOUT, fow.ROUNDABOUT)] = EXCELLENT
_FOWRating[_to_index(fow.ROUNDABOUT, fow.TRAFFICSQUARE)] = AVERAGE
_FOWRating[_to_index(fow.ROUNDABOUT, fow.SLIPROAD)] = POOR
_FOWRating[_to_index(fow.ROUNDABOUT, fow.OTHER)] = POOR

_FOWRating[_to_index(fow.TRAFFICSQUARE, fow.TRAFFICSQUARE)] = EXCELLENT
_FOWRating[_to_index(fow.TRAFFICSQUARE, fow.SLIPROAD)] = POOR
_FOWRating[_to_index(fow.TRAFFICSQUARE, fow.OTHER)] = POOR

_FOWRating[_to_index(fow.SLIPROAD, fow.SLIPROAD)] = EXCELLENT
_FOWRating[_to_index(fow.SLIPROAD, fow.OTHER)] = POOR

_FOWRating[_to_index(fow.OTHER, fow.OTHER)] = EXCELLENT


def get_fow_rating_category(fow1, fow2):
    """ Get FOW rating category
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Polygon location
        :rtype: PolygonLocation
    """
    if fow1 > fow2:
        fow2, fow1 = fow1, fow2
    return _FOWRating[_to_index(fow1, fow2)]



# -*- coding: utf-8 -*-
''' Constante definitions

    .. moduleauthor:: David Marteau <david.marteau@mappy.com>
'''


from gv_vendor.pylr.utils import enum

''' The data format IDENTIFIER. '''
IDENTIFIER = "binary"


''' The Constant BITS_PER_BYTE. '''
BITS_PER_BYTE = 8

'''
The Constant DECA_MICRO_DEG_FACTOR is used to transform degree values
into deca-micro degrees.
'''
DECA_MICRO_DEG_FACTOR = 100000.0

'''
The Constant BIT24FACTOR is used for the conversion of lat/lon
coordinates into a 24bit accuracy.
'''
BIT24FACTOR = 46603.377778

'''
The Constant BIT24FACTOR_REVERSED is used for the conversion of 24bit
lat/lon values back into prior accuracy.
'''
BIT24FACTOR_REVERSED = 1 / BIT24FACTOR

''' The BEARING_SECTOR defines the length of a bearing interval. '''
BEARING_SECTOR = 11.25

''' The LENGTH_INTERVAL defines the length of a dnp and offset interval. '''
LENGTH_INTERVAL = 58.6

''' The IS_POINT defines a polocation reference. '''
IS_POINT = 1

''' The IS_NOT_POINT indicates that the location reference is not a polocation. '''
IS_NOT_POINT = 0

''' The IS_AREA defines an area location reference. '''
IS_AREA = 1

'''  The AREA_CODE_CIRCLE defines the code for a cirle location reference. '''
AREA_CODE_CIRCLE = 0

'''  The AREA_CODE_RECTANGLE defines the code for a rectangle location reference. '''
AREA_CODE_RECTANGLE = 2

'''  The AREA_CODE_GRID defines the code for a grid location reference. '''
AREA_CODE_GRID = 2  # for BINARY_VERSION_3 the same as for AREA_CODE_RECTANGLE

'''  The AREA_CODE_POLYGON defines the code for a polygon location reference. '''
AREA_CODE_POLYGON = 1

'''  The AREA_CODE_CLOSEDLINE defines the code for a closed line location reference. '''
AREA_CODE_CLOSEDLINE = 3

'''  The AREA_CODE_NOAREA defines the code for a non-area location reference. '''
IS_NOT_AREA = 0

''' The HAS_ATTRIBUTES the existence of attribute information in the stream. '''
HAS_ATTRIBUTES = 1

''' The Constant HAS_NO_ATTRIBUTES. '''
HAS_NO_ATTRIBUTES = 0

''' The HEADER_SIZE defines the size [in bytes] of the header. '''
HEADER_SIZE = 1

'''
The FIRST_LRP_SIZE defines the size [in bytes] of the first location
reference point.
'''
FIRST_LRP_SIZE = 9

'''
The LRP_SIZE defines the size [in bytes] of an intermediate location
reference point.
'''
LRP_SIZE = 7

'''
The LAST_LRP_SIZE defines the size [in bytes] of the last location
reference point.
'''
LAST_LRP_SIZE = 6

''' The Constant ABS_COORD_SIZE. '''
ABS_COORD_SIZE = 6

''' The Constant RELATIVE_OFFSET_LENGTH. '''
RELATIVE_OFFSET_LENGTH = 0.390625

'''
The MIN_BYTES defines the minimum size [in bytes] of a binary location
reference.
'''
MIN_BYTES_LINE_LOCATION = HEADER_SIZE + FIRST_LRP_SIZE + LAST_LRP_SIZE

'''
The MIN_BYTES defines the minimum size [in bytes] of a binary closed line location
reference.
'''
MIN_BYTES_CLOSED_LINE_LOCATION = HEADER_SIZE + FIRST_LRP_SIZE + 2

''' The Constant GEOCOORD_SIZE. '''
GEOCOORD_SIZE = HEADER_SIZE + ABS_COORD_SIZE

''' The Constant MIN_BYTES_POINT_LOCATION. '''
MIN_BYTES_POINT_LOCATION = GEOCOORD_SIZE

''' The Constant BINARY_VERSION_2. '''
BINARY_VERSION_2 = 2

''' The Constant BINARY_VERSION_3. '''
BINARY_VERSION_3 = 3

''' The LATEST_BINARY_VERSION defines the current version of the binary format. '''
LATEST_BINARY_VERSION = BINARY_VERSION_3

''' The HAS_OFFSET defines the existence of offset information. '''
HAS_OFFSET = 1

''' The Constant OFFSET_BUCKETS. '''
OFFSET_BUCKETS = 256

''' The Constant POINT_ALONG_LINE_SIZE. '''
POINT_ALONG_LINE_SIZE = HEADER_SIZE + FIRST_LRP_SIZE + LAST_LRP_SIZE

''' The Constant RELATIVE_COORD_SIZE. '''
RELATIVE_COORD_SIZE = 4

''' number of bits used for a small radius '''
SMALL_RADIUS_BITS = 8

''' number of bits used for a medium radius '''
MEDIUM_RADIUS_BITS = 16

''' number of bits used for a large radius '''
LARGE_RADIUS_BITS = 24

''' number of bits used for a small radius '''
EXTRA_LARGE_RADIUS_BITS = 32

''' The Constant DIMENSION_SIZE. '''
DIMENSION_SIZE = 2

''' The Constant RECTANGLE_SIZE. '''
RECTANGLE_SIZE = HEADER_SIZE + ABS_COORD_SIZE + RELATIVE_COORD_SIZE

''' The Constant LARGE_RECTANGLE_SIZE. '''
LARGE_RECTANGLE_SIZE = HEADER_SIZE + ABS_COORD_SIZE + ABS_COORD_SIZE

''' The Constant GRID_SIZE. '''
GRID_SIZE = RECTANGLE_SIZE + 2 * DIMENSION_SIZE

''' The Constant LARGE_GRID_SIZE. '''
LARGE_GRID_SIZE = LARGE_RECTANGLE_SIZE + 2 * DIMENSION_SIZE

''' The Constant MIN_BYTES_POLYGON. '''
MIN_BYTES_POLYGON = HEADER_SIZE + ABS_COORD_SIZE + 2 * RELATIVE_COORD_SIZE

''' The Constant POINT_OFFSET_SIZE. '''
POINT_OFFSET_SIZE = 1

''' The Constant POINT_WITH_ACCESS_SIZE. '''
POINT_WITH_ACCESS_SIZE = HEADER_SIZE + FIRST_LRP_SIZE + LAST_LRP_SIZE + RELATIVE_COORD_SIZE

''' The Constant POINT_LOCATION_VERSION. '''
POINT_LOCATION_VERSION = 3

''' Base size for circle location '''
CIRCLE_BASE_SIZE = HEADER_SIZE + ABS_COORD_SIZE

LocationType = enum(
    # Location is UNKNOWN.
    'UNKNOWN',

    # line location.
    'LINE_LOCATION',

    # simple geo coordinates
    'GEO_COORDINATES',

    # point along a line
    'POINT_ALONG_LINE',

    # point of interest with an access point along a line
    'POI_WITH_ACCESS_POINT',

    # circle area location
    'CIRCLE',

    # polygon area location
    'POLYGON',

    # closed line area location
    'CLOSED_LINE',

    # rectangular area location
    'RECTANGLE',

    # grid area location
    'GRID'
    )


''' The Constant POINT_LOCATION_TYPES. '''
POINT_LOCATION_TYPES = {LocationType.GEO_COORDINATES, LocationType.POI_WITH_ACCESS_POINT, LocationType.POINT_ALONG_LINE}

''' The Constant AREA_LOCATION_VERSION. '''
AREA_LOCATION_VERSION = 3

''' The Constant AREA_LOCATION_TYPES. '''
AREA_LOCATION_TYPES = {LocationType.CIRCLE, LocationType.GRID, LocationType.CLOSED_LINE, LocationType.RECTANGLE, LocationType.POLYGON}

''' Side constants '''
ON_ROAD_OR_UNKNOWN = 0
RIGTH_SIDE = 1
LEFT_SIDE = 2
BOTH_SIDE = RIGTH_SIDE | LEFT_SIDE

''' Point has no sense of orientation, or determination of
    orientation is not applicable
'''
NO_ORIENTATION_OR_UNKNOWN = 0

''' Point has orientation from first LRP towards second LRP '''
WITH_LINE_DIRECTION = 1

''' Point has orientation from second LRP towards first LRP '''
AGAINST_LINE_DIRECTION = 2

''' Point has orientation in both directions '''
BOTH_DIRECTIONS = WITH_LINE_DIRECTION | AGAINST_LINE_DIRECTION



# -*- coding: utf-8 -*-
''' Define functions for transforming encoded values

    .. moduleauthor:: David Marteau <david.marteau@mappy.com>

'''

from gv_vendor.pylr.utils import signum
from gv_vendor.pylr.constants import (BIT24FACTOR_REVERSED,
                                      DECA_MICRO_DEG_FACTOR,
                                      BEARING_SECTOR,
                                      LENGTH_INTERVAL,
                                      RELATIVE_OFFSET_LENGTH)


def _get32BitRepresentation(v):
    """ Calculate the 32 bit double value representation of a coordinate
        out of a 24 bit integer value representation.
        
        :param int v: Coordinate expressed in 24 bit integer value
        :return: Coordinate (in degrees)
        :rtype: float  
    """
    return (v - signum(v)*0.5) * BIT24FACTOR_REVERSED


def coordinates_values(lon, lat):
    """ Calculate the 32 bit double value representations of a pair of coordinates
        out of 24 bit integer values representation.
        
        :param int lon: Longitude expressed in 24 bit integer value
        :param int lat: Latitude expressed in 24 bit integer value
        :return: Pair of coordinates (in degrees)
        :rtype: float  
    """
    return (_get32BitRepresentation(lon),
            _get32BitRepresentation(lat))


def rel_coordinates_values(prev, blon, blat ):
    """ Calculate absolute corrdinates from relative coordinates to
        apply to the absolute coordinates of a reference position.
        
        :param Coords prev: Absolute coordinates of the reference geo point
        :param int blon: relative longitude expressed in decamicrodegrees
        :param int blat: relative latitude expressed in decamicrodegrees
        :return: Pair of coordinates (in degrees)
        :rtype: float
    """
    lon = (prev.lon + blon / DECA_MICRO_DEG_FACTOR)
    lat = (prev.lat + blat / DECA_MICRO_DEG_FACTOR)
    return lon, lat


def bearing_estimate(interval):
    """ Calculates an estimate for the bearing value. The bearing information
        provided by the location reference point indicates an interval in which
        the concrete value is. The approximation is the middle of that interval.
        
        :param int interval: bearing interval
        :return: bearing estimation (in degrees)
        :rtype: float
        
    """
    lower = interval * BEARING_SECTOR
    upper = (interval + 1) * BEARING_SECTOR
    return ((upper + lower) / 2)


def distance_estimate(interval):
    """ Calculates an estimate for a distance value. The distance information
        provided by the location reference point indicates an interval in which
        the concrete value is. The approximation is the middle of that interval.

        :param int interval: distance interval
        :return: distance estimation (in meters)
        :rtype: int
    """
    lower = interval * LENGTH_INTERVAL
    upper = (interval + 1) * LENGTH_INTERVAL
    return round(((upper + lower) / 2))


def relative_distance(offset):
    """ Calculates an estimate for a relative offset value. The offset information
        provided by the location reference point indicates an interval in which
        the concrete value (percentage) is. The approximation is the middle of
        that interval.

        :param int offset: offset interval
        :return: offset estimation (in %)
        :rtype: float
    """
    lower = offset * RELATIVE_OFFSET_LENGTH
    upper = (offset + 1) * RELATIVE_OFFSET_LENGTH
    return (lower + upper) / 2



#!/usr/bin/env python3

import aiohttp

from gv_utils import parser
from gv_utils.datetime import datetime, to_str
from gv_utils.enums import RequestParam, RequestPath
from gv_utils.geometry import BaseGeometry, encode_geometry
from gv_utils.logger import Logger

SEPARATOR = RequestPath.separator


class Requests:

    def __init__(self, logger: Logger) -> None:
        self.logger = logger
        self.apiurl = None
        self.session = None

    async def async_init(self, apiaddr):
        self.apiurl = 'http://{}/api'.format(apiaddr)
        self.session = aiohttp.ClientSession()

    async def get_data_point(self, datapointeids: list = None, datatypeeids: list = None,
                             area: BaseGeometry = None) -> dict:
        params = {}
        if datatypeeids is not None:
            params[RequestParam.datatype] = self.__encode_list_params(datatypeeids)
        return await self._get_location(datapointeids, area, RequestPath.datapoint, params)

    async def get_road(self, roadeids: list = None, area: BaseGeometry = None) -> dict:
        return await self._get_location(roadeids, area, RequestPath.road)

    async def get_zone_point(self, zoneeids: list = None, area: BaseGeometry = None) -> dict:
        return await self._get_location(zoneeids, area, RequestPath.zone)

    async def _get_location(self, eids: list, area: BaseGeometry, applyto: str, params: dict = None) -> dict:
        if params is None:
            params = {}

        if eids is not None:
            params[RequestParam.eid] = self.__encode_list_params(eids)
        if area is not None:
            params[RequestParam.within] = encode_geometry(area)
        return await self._get(applyto, params)

    async def get_road_data_point(self, area: BaseGeometry = None) -> dict:
        params = {}
        if area is not None:
            params[RequestParam.within] = encode_geometry(area)
        return await self._get(SEPARATOR.join((RequestPath.road, RequestPath.datapoint)), params)

    async def get_data_point_data_quality(self, period: str, fromdatetime: datetime, datapointeids: list = None,
                                          datatypeeids: list = None, todatetime: datetime = None) -> dict:
        params = {}
        if datatypeeids is not None:
            params[RequestParam.datatype] = self.__encode_list_params(datatypeeids)
        return await self._get_data_quality(period, fromdatetime, datapointeids, todatetime, RequestPath.datapoint,
                                            params)

    async def get_data_type_data_quality(self, period: str, fromdatetime: datetime, datatypeeids: list = None,
                                         todatetime: datetime = None) -> dict:
        return await self._get_data_quality(period, fromdatetime, datatypeeids, todatetime, RequestPath.datatype)

    async def get_road_data_quality(self, period: str, fromdatetime: datetime, roadeids: list = None,
                                    todatetime: datetime = None) -> dict:
        return await self._get_data_quality(period, fromdatetime, roadeids, todatetime, RequestPath.road)

    async def _get_data_quality(self, period: str, fromdatetime: datetime, locationeids: list, todatetime: datetime,
                                applyto: str, params: dict = None) -> dict:
        return await self.__get_data(period, fromdatetime, locationeids, todatetime,
                                     SEPARATOR.join((applyto, RequestPath.dataquality)), params)

    async def get_data_point_indicator(self, indicator: str, period: str, fromdatetime: datetime,
                                       datapointeids: list = None, datatypeeids: list = None,
                                       todatetime: datetime = None) -> dict:
        params = {}
        if datatypeeids is not None:
            params[RequestParam.datatype] = self.__encode_list_params(datatypeeids)
        return await self._get_indicator(indicator, period, fromdatetime, datapointeids, todatetime,
                                         RequestPath.datapoint, params)

    async def get_data_point_imputed_indicator(self, indicator: str, period: str, fromdatetime: datetime,
                                               datapointeids: list = None, datatypeeids: list = None,
                                               todatetime: datetime = None) -> dict:
        params = {}
        if datatypeeids is not None:
            params[RequestParam.datatype] = self.__encode_list_params(datatypeeids)
        return await self._get_indicator(indicator, period, fromdatetime, datapointeids, todatetime,
                                         SEPARATOR.join((RequestPath.datapoint, RequestPath.imputed)), params)

    async def get_road_indicator(self, indicator: str, period: str, fromdatetime: datetime, roadeids: list = None,
                                 todatetime: datetime = None) -> dict:
        return await self._get_indicator(indicator, period, fromdatetime, roadeids, todatetime, RequestPath.road)

    async def _get_indicator(self, indicator: str, period: str, fromdatetime: datetime, locationeids: list,
                             todatetime: datetime, applyto: str, params: dict = None) -> dict:
        if params is None:
            params = {}

        params[RequestParam.indicator] = indicator
        return await self.__get_data(period, fromdatetime, locationeids, todatetime,
                                     SEPARATOR.join((applyto, RequestPath.indicator)), params)

    async def __get_data(self, period: str, fromdatetime: datetime, locationeids: list, todatetime: datetime,
                         applyto: str, params) -> dict:
        if params is None:
            params = {}

        params[RequestParam.period] = period
        params[RequestParam.fromdatetime] = to_str(fromdatetime)
        if locationeids is not None:
            params[RequestParam.eid] = self.__encode_list_params(locationeids)
        if todatetime is not None:
            params[RequestParam.todatetime] = to_str(todatetime)
        return await self._get(applyto, params)

    async def get_zone_travel_time(self, frompointeid: str, topointeid: str, period: str, fromdatetime: datetime,
                                   todatetime: datetime = None) -> dict:
        params = {
            RequestParam.frompoint: frompointeid,
            RequestParam.topoint: topointeid,
            RequestParam.period: period,
            RequestParam.fromdatetime: to_str(fromdatetime)
        }
        if todatetime is not None:
            params[RequestParam.todatetime] = to_str(todatetime)
        return await self._get(SEPARATOR.join((RequestPath.zonepoint, RequestPath.traveltime)), params)

    async def _get(self, suffix: str, params: dict = None) -> dict:
        message = {}
        async with self.session as session:
            url = '{}{}'.format(SEPARATOR.join((self.apiurl, suffix)), SEPARATOR)
            async with session.get(url, params=params) as resp:
                if resp.status == 200:
                    message = await parser.decode_message(await resp.text())
        return message

    @staticmethod
    def __encode_list_params(params: list) -> str:
        return RequestParam.separator.join(params)

    async def close(self) -> None:
        await self.session.close()



# -*- coding: utf-8 -*-
''' Support definition and intermediate parsing of binary encoded locations

.. moduleauthor:: David Marteau <david.marteau@mappy.com>

'''

from collections import namedtuple
from gv_vendor.pylr.values import (coordinates_values,
                                   rel_coordinates_values,
                                   distance_estimate,
                                   relative_distance)

from gv_vendor.pylr.constants import (BINARY_VERSION_2, BINARY_VERSION_3)

# lon (float) : longitude (in degrees)
# lat (float) : latituded (in degrees)
Coords = namedtuple('Coords',  ('lon', 'lat'))
"""Coordinates, expressed as a pair of float values (in degrees)."""

# coords (Coords) : Coordinates
# bear : bearing ; expressed as a sector number describing a bearing range
# orient : orientation, expressed as a bit field
# frc : functionnal road class, expressed as a bit field
# fow : form of way, expressed as a bit field
# lfrcnp : lowest FRC to next point, expressed as a bit field
# dnp : distance to next LR-point, expressed in meters
LocationReferencePoint = namedtuple('LocationReferencePoint',  (
    'coords', 'bear', 'orient', 'frc', 'fow',  'lfrcnp', 'dnp'))
"Location Reference Point. Used to define several kinds of lcoation (lines, ...)"

RADIUS_BITS = ('uint:8', 'uint:16', 'uint:24', 'uint:32')
""" Number of bits needed to code the radius for a circle location reference.
"""

# Bit fields used to read physical data
RFU_BITS = 'uint:1'
SIDE_OR_ORIENTATION_BITS = 'uint:2' 
FRC_BITS = 'uint:3'
FOW_BITS = 'uint:3'
LFRCNP_BITS = 'uint:3'
BEAR_BITS = 'uint:5'
DNP_BITS = 'uint:8'
POFFF_BITS = 'uint:1'
NOFFF_BITS = 'uint:1'
NR_RFU2 = 'uint:2'
NR_RFU3 = 'uint:3'
GRID_CELL_BITS = 'uint:16'
ABS_COORDS_BITS = 'int:24'
RLRP_COORDS_BITS = 'int:16'
OFFSET_BITS = 'uint:8'

# Caution : the attribute names refer to the reference implementation,
# not to the specifications
ATTR1_BITS = (SIDE_OR_ORIENTATION_BITS, FRC_BITS, FOW_BITS)
ATTR2_BITS = (LFRCNP_BITS, BEAR_BITS)
ATTR3_BITS = (DNP_BITS,)
ATTR4_BITS = (RFU_BITS, POFFF_BITS, NOFFF_BITS, BEAR_BITS)
ATTR5_BITS = (NR_RFU2, FRC_BITS, FOW_BITS)
ATTR6_BITS = (NR_RFU3, BEAR_BITS)

def _parse_attr1(rb):
    ''' Parse the raw binary data bits defined by ATTR1_BITS
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: orientation, FRC, FOW
        :rtype: int, int, int
    '''
    orient, frc, fow = rb.getbits(*ATTR1_BITS)
    return orient, frc, fow


def _parse_attr2(rb):
    ''' Parse the raw binary data bits defined by ATTR2_BITS
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: lowest FRC to next LR-point, bearing
        :rtype: int, int
    '''
    lfrcnp, bear = rb.getbits(*ATTR2_BITS)
    return lfrcnp, bear


def _parse_attr3(rb):
    ''' Parse the raw binary data bits defined by ATTR3_BITS
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: distance interval to next LR-point
        :rtype: int
    '''
    dnp = rb.getbits(*ATTR3_BITS)
    return dnp


def _parse_attr4(rb):
    ''' Parse the raw binary data bits defined by ATTR3_BITS
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: positive offset length interval, negative offset length interval, bearing sector
        :rtype: int, int, int
    '''
    _, pofff, nofff, bear = rb.getbits(*ATTR4_BITS)
    return pofff, nofff, bear


def _parse_attr5(rb):
    ''' Parse the raw binary data bits defined by ATTR3_BITS
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Functionnal Road Class, Form Of Way
        :rtype: int, int
    '''
    _, frc, fow = rb.getbits(*ATTR5_BITS)
    return frc, fow


def _parse_attr6(rb):
    ''' Parse the raw binary data bits defined by ATTR3_BITS
        :param _RawBinaryData rb: Binary data describing the location
        :returns: bearing sector
        :rtype: int
    '''
    _, bear = rb.getbits(*ATTR6_BITS)
    return bear


def _parse_coordinates(rb, bits, rel):
    ''' Parse the raw binary data bits to read coordinates. These ones can be read as absolute or relative coordinates.
        In the last case, the third parameter must be provided. Else, its value must be None.
        
        :param _RawBinaryData rb: Binary data describing the location
        :param string bits: String describing the number of bits to read ; may be RLRP_COORDS_BITS or ABS_COORDS_BITS.
        :param Coords rel: If the ccordinates to read are supposed to be relative, this parameter gives the reference coordinates.
                           Its value is ignored when asbulute coordinates is set for the second parameter.
        :returns: Coordinates
        :rtype: Coords
    '''
    lon, lat = rb.getbits(bits, bits)
    if rel:
        lon, lat = rel_coordinates_values(rel, lon, lat)
    else:
        lon, lat = coordinates_values(lon, lat)
    return Coords(lon, lat)

def _parse_relative_coordinates(rb, rel):
    ''' Parse the raw binary data bits to read relative coordinates.
    
        :param _RawBinaryData rb: Binary data describing the location
        :param Coords rel: Reference coordinates
        :returns: Coordinates
        :rtype: Coords
    '''
    return _parse_coordinates(rb, RLRP_COORDS_BITS, rel)


def _parse_absolute_coordinates(rb):
    ''' Parse the raw binary data bits to read absolute coordinates.
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Coordinates
        :rtype: Coords
    '''
    return _parse_coordinates(rb, ABS_COORDS_BITS, None)


def _parse_lrp(rb, bits, rel_coords):
    ''' Parse location reference point
    
        :param _RawBinaryData rb: Binary data describing the location
        :param string bits: String describing the number of bits to read ; may be RLRP_COORDS_BITS or ABS_COORDS_BITS.
        :param Coords rel_coords: Reference coordinates.
        :returns: Location Reference Point
        :rtype: LocationReferencePoint
    '''
    coords = _parse_coordinates(rb, bits, rel_coords)
    orient, frc, fow = _parse_attr1(rb)
    frcnp, bear = _parse_attr2(rb)
    dnp, = _parse_attr3(rb)

    dnp = distance_estimate(dnp)

    return LocationReferencePoint(coords, bear, orient, frc, fow, frcnp, dnp)


def _parse_first_lrp(rb):
    ''' Parse the first location reference point (coded in absolute coordinates) of a location reference
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Location Reference Point
        :rtype: LocationReferencePoint
    '''
    return _parse_lrp(rb, ABS_COORDS_BITS, None)


def _parse_intermediate_lrp(rb, rel):
    ''' Parse any intermediate location reference point (coded in relative coordinates) of a location reference
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Location Reference Point
        :rtype: LocationReferencePoint
    '''
    return _parse_lrp(rb, RLRP_COORDS_BITS, rel.coords)


def _parse_last_line_lrp(rb, rel, sided=False):
    ''' Parse the last location reference point for some kinds of location references
    
        :param _RawBinaryData rb: Binary data describing the location
        :param Coords rel: Reference coordinates
        :param bool sided: Must be set to True if a diding info must be read
        :returns: Location Reference Point, positive offset length interval, negative offset length interval
        :rtype: LocationReferencePoint, int, int
    '''
    coords = _parse_coordinates(rb, RLRP_COORDS_BITS, rel.coords)
    side, frc, fow = _parse_attr1(rb)
    pofff, nofff, bear = _parse_attr4(rb)

    return LocationReferencePoint(coords, bear, side, frc, fow, None, None), pofff, nofff


def _parse_last_closed_line_attrs(rb):
    ''' Parse the last location reference point's attributes for the closed line location reference
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Functionnal Road Class, Form Of Way, bearing sector
        :rtype: int, int, int
    '''
    frc, fow = _parse_attr5(rb)
    bear = _parse_attr6(rb)

    return frc, fow, bear


def _parse_offset(rb):
    ''' Parse positive or negative offset for some kinds of location references
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Estimate distance, in meters
        :rtype: int
    '''
    offs, = rb.getbits(OFFSET_BITS)
    if rb.version == BINARY_VERSION_2:
        offs = distance_estimate(offs)
    elif rb.version == BINARY_VERSION_3:
        offs = relative_distance(offs)
    return offs


def _parse_radius(rb, radius_size):
    ''' Parse radius of circle location references.
    
        :param _RawBinaryData rb: Binary data describing the location
        :param int radius_size: Number of bytes needed to read the radius
        :returns: Radius, in meters
        :rtype: int
    '''
    radius, = rb.getbits(RADIUS_BITS[radius_size-1])
    return radius


def _parse_grid_dimensions(rb):
    ''' Parse grid location reference dimensions
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Number of columns, number of rows
        :rtype: int, int
    '''
    cols, = rb.getbits(GRID_CELL_BITS)
    rows, = rb.getbits(GRID_CELL_BITS)
    return cols, rows



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ProtobufTrafficFlow_v5.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ProtobufTrafficFlow_v5.proto',
  package='trafficFlow',
  syntax='proto2',
  serialized_pb=_b('\n\x1cProtobufTrafficFlow_v5.proto\x12\x0btrafficFlow\"\xa2\x02\n\x10TrafficFlowGroup\x12\x35\n\x0fmetaInformation\x18\x01 \x01(\x0b\x32\x1c.trafficFlow.MetaInformation\x12-\n\x0btrafficFlow\x18\x02 \x03(\x0b\x32\x18.trafficFlow.TrafficFlow\x12I\n\x19trafficFlowWithPrediction\x18\x03 \x03(\x0b\x32&.trafficFlow.TrafficFlowWithPrediction\x12]\n#trafficFlowWithPredictionPerSection\x18\x04 \x03(\x0b\x32\x30.trafficFlow.TrafficFlowWithPredictionPerSection\"\x89\x01\n\x0fMetaInformation\x12\x1c\n\x14\x63reateTimeUTCSeconds\x18\x01 \x01(\r\x12\x41\n\x15supplierAndClientInfo\x18\x02 \x01(\x0b\x32\".trafficFlow.SupplierAndClientInfo\x12\x15\n\rformatVersion\x18\x03 \x01(\x05\"=\n\x15SupplierAndClientInfo\x12\x10\n\x08\x63lientID\x18\x01 \x01(\t\x12\x12\n\nsupplierID\x18\x02 \x01(\t\"u\n\x0bTrafficFlow\x12\'\n\x08location\x18\x01 \x01(\x0b\x32\x15.trafficFlow.Location\x12!\n\x05speed\x18\x02 \x01(\x0b\x32\x12.trafficFlow.Speed\x12\x1a\n\x0broadClosure\x18\x03 \x01(\x08:\x05\x66\x61lse\"\xae\x01\n\x19TrafficFlowWithPrediction\x12\'\n\x08location\x18\x01 \x01(\x0b\x32\x15.trafficFlow.Location\x12.\n\x05speed\x18\x02 \x03(\x0b\x32\x1f.trafficFlow.SpeedWithTimeStamp\x12\x1a\n\x0broadClosure\x18\x03 \x01(\x08:\x05\x66\x61lse\x12\x1c\n\x14timeToUsualInMinutes\x18\x04 \x01(\r\"\x9b\x01\n#TrafficFlowWithPredictionPerSection\x12\'\n\x08location\x18\x01 \x01(\x0b\x32\x15.trafficFlow.Location\x12-\n\x0bspeedVector\x18\x02 \x03(\x0b\x32\x18.trafficFlow.SpeedVector\x12\x1c\n\x14timeToUsualInMinutes\x18\x03 \x01(\r\"?\n\x08Location\x12\x0b\n\x03tmc\x18\x01 \x01(\t\x12\x0e\n\x06openlr\x18\x02 \x01(\x0c\x12\x16\n\x0elengthInMeters\x18\x03 \x01(\r\"\xb9\x01\n\x05Speed\x12\x18\n\x10\x61verageSpeedKmph\x18\x01 \x01(\x05\x12\x19\n\x11travelTimeSeconds\x18\x02 \x01(\x05\x12\x12\n\nconfidence\x18\x03 \x01(\x05\x12\x15\n\rrelativeSpeed\x18\x04 \x01(\x02\x12\x37\n\x10trafficCondition\x18\x05 \x01(\x0e\x32\x1d.trafficFlow.TrafficCondition\x12\x17\n\x0f\x61verageSpeedMph\x18\x06 \x01(\x05\"P\n\x12SpeedWithTimeStamp\x12\x17\n\x0fminutesInFuture\x18\x01 \x01(\r\x12!\n\x05speed\x18\x02 \x01(\x0b\x32\x12.trafficFlow.Speed\"W\n\x0bSpeedVector\x12\x17\n\x0fminutesInFuture\x18\x01 \x01(\r\x12/\n\x0csectionSpeed\x18\x02 \x03(\x0b\x32\x19.trafficFlow.SectionSpeed\"j\n\x0cSectionSpeed\x12\x1b\n\x13startOffsetInMeters\x18\x01 \x01(\r\x12!\n\x05speed\x18\x02 \x01(\x0b\x32\x12.trafficFlow.Speed\x12\x1a\n\x0broadClosure\x18\x03 \x01(\x08:\x05\x66\x61lse*\xa8\x01\n\x10TrafficCondition\x12\x17\n\x13LONG_QUEUES_TRAFFIC\x10\x01\x12\x10\n\x0c\x46REE_TRAFFIC\x10\x02\x12\x11\n\rHEAVY_TRAFFIC\x10\x03\x12\x10\n\x0cSLOW_TRAFFIC\x10\x04\x12\x13\n\x0fQUEUING_TRAFFIC\x10\x05\x12\x16\n\x12STATIONARY_TRAFFIC\x10\x06\x12\n\n\x06\x43LOSED\x10\x07\x12\x0b\n\x07UNKNOWN\x10\x08')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_TRAFFICCONDITION = _descriptor.EnumDescriptor(
  name='TrafficCondition',
  full_name='trafficFlow.TrafficCondition',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LONG_QUEUES_TRAFFIC', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FREE_TRAFFIC', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HEAVY_TRAFFIC', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SLOW_TRAFFIC', index=3, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='QUEUING_TRAFFIC', index=4, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STATIONARY_TRAFFIC', index=5, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLOSED', index=6, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=7, number=8,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1528,
  serialized_end=1696,
)
_sym_db.RegisterEnumDescriptor(_TRAFFICCONDITION)

TrafficCondition = enum_type_wrapper.EnumTypeWrapper(_TRAFFICCONDITION)
LONG_QUEUES_TRAFFIC = 1
FREE_TRAFFIC = 2
HEAVY_TRAFFIC = 3
SLOW_TRAFFIC = 4
QUEUING_TRAFFIC = 5
STATIONARY_TRAFFIC = 6
CLOSED = 7
UNKNOWN = 8



_TRAFFICFLOWGROUP = _descriptor.Descriptor(
  name='TrafficFlowGroup',
  full_name='trafficFlow.TrafficFlowGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='metaInformation', full_name='trafficFlow.TrafficFlowGroup.metaInformation', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='trafficFlow', full_name='trafficFlow.TrafficFlowGroup.trafficFlow', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='trafficFlowWithPrediction', full_name='trafficFlow.TrafficFlowGroup.trafficFlowWithPrediction', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='trafficFlowWithPredictionPerSection', full_name='trafficFlow.TrafficFlowGroup.trafficFlowWithPredictionPerSection', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=46,
  serialized_end=336,
)


_METAINFORMATION = _descriptor.Descriptor(
  name='MetaInformation',
  full_name='trafficFlow.MetaInformation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='createTimeUTCSeconds', full_name='trafficFlow.MetaInformation.createTimeUTCSeconds', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='supplierAndClientInfo', full_name='trafficFlow.MetaInformation.supplierAndClientInfo', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='formatVersion', full_name='trafficFlow.MetaInformation.formatVersion', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=339,
  serialized_end=476,
)


_SUPPLIERANDCLIENTINFO = _descriptor.Descriptor(
  name='SupplierAndClientInfo',
  full_name='trafficFlow.SupplierAndClientInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='clientID', full_name='trafficFlow.SupplierAndClientInfo.clientID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='supplierID', full_name='trafficFlow.SupplierAndClientInfo.supplierID', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=478,
  serialized_end=539,
)


_TRAFFICFLOW = _descriptor.Descriptor(
  name='TrafficFlow',
  full_name='trafficFlow.TrafficFlow',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='location', full_name='trafficFlow.TrafficFlow.location', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='speed', full_name='trafficFlow.TrafficFlow.speed', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='roadClosure', full_name='trafficFlow.TrafficFlow.roadClosure', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=541,
  serialized_end=658,
)


_TRAFFICFLOWWITHPREDICTION = _descriptor.Descriptor(
  name='TrafficFlowWithPrediction',
  full_name='trafficFlow.TrafficFlowWithPrediction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='location', full_name='trafficFlow.TrafficFlowWithPrediction.location', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='speed', full_name='trafficFlow.TrafficFlowWithPrediction.speed', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='roadClosure', full_name='trafficFlow.TrafficFlowWithPrediction.roadClosure', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timeToUsualInMinutes', full_name='trafficFlow.TrafficFlowWithPrediction.timeToUsualInMinutes', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=661,
  serialized_end=835,
)


_TRAFFICFLOWWITHPREDICTIONPERSECTION = _descriptor.Descriptor(
  name='TrafficFlowWithPredictionPerSection',
  full_name='trafficFlow.TrafficFlowWithPredictionPerSection',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='location', full_name='trafficFlow.TrafficFlowWithPredictionPerSection.location', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='speedVector', full_name='trafficFlow.TrafficFlowWithPredictionPerSection.speedVector', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timeToUsualInMinutes', full_name='trafficFlow.TrafficFlowWithPredictionPerSection.timeToUsualInMinutes', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=838,
  serialized_end=993,
)


_LOCATION = _descriptor.Descriptor(
  name='Location',
  full_name='trafficFlow.Location',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tmc', full_name='trafficFlow.Location.tmc', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='openlr', full_name='trafficFlow.Location.openlr', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='lengthInMeters', full_name='trafficFlow.Location.lengthInMeters', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=995,
  serialized_end=1058,
)


_SPEED = _descriptor.Descriptor(
  name='Speed',
  full_name='trafficFlow.Speed',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='averageSpeedKmph', full_name='trafficFlow.Speed.averageSpeedKmph', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='travelTimeSeconds', full_name='trafficFlow.Speed.travelTimeSeconds', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='confidence', full_name='trafficFlow.Speed.confidence', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='relativeSpeed', full_name='trafficFlow.Speed.relativeSpeed', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='trafficCondition', full_name='trafficFlow.Speed.trafficCondition', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='averageSpeedMph', full_name='trafficFlow.Speed.averageSpeedMph', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1061,
  serialized_end=1246,
)


_SPEEDWITHTIMESTAMP = _descriptor.Descriptor(
  name='SpeedWithTimeStamp',
  full_name='trafficFlow.SpeedWithTimeStamp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='minutesInFuture', full_name='trafficFlow.SpeedWithTimeStamp.minutesInFuture', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='speed', full_name='trafficFlow.SpeedWithTimeStamp.speed', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1248,
  serialized_end=1328,
)


_SPEEDVECTOR = _descriptor.Descriptor(
  name='SpeedVector',
  full_name='trafficFlow.SpeedVector',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='minutesInFuture', full_name='trafficFlow.SpeedVector.minutesInFuture', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sectionSpeed', full_name='trafficFlow.SpeedVector.sectionSpeed', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1330,
  serialized_end=1417,
)


_SECTIONSPEED = _descriptor.Descriptor(
  name='SectionSpeed',
  full_name='trafficFlow.SectionSpeed',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='startOffsetInMeters', full_name='trafficFlow.SectionSpeed.startOffsetInMeters', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='speed', full_name='trafficFlow.SectionSpeed.speed', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='roadClosure', full_name='trafficFlow.SectionSpeed.roadClosure', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1419,
  serialized_end=1525,
)

_TRAFFICFLOWGROUP.fields_by_name['metaInformation'].message_type = _METAINFORMATION
_TRAFFICFLOWGROUP.fields_by_name['trafficFlow'].message_type = _TRAFFICFLOW
_TRAFFICFLOWGROUP.fields_by_name['trafficFlowWithPrediction'].message_type = _TRAFFICFLOWWITHPREDICTION
_TRAFFICFLOWGROUP.fields_by_name['trafficFlowWithPredictionPerSection'].message_type = _TRAFFICFLOWWITHPREDICTIONPERSECTION
_METAINFORMATION.fields_by_name['supplierAndClientInfo'].message_type = _SUPPLIERANDCLIENTINFO
_TRAFFICFLOW.fields_by_name['location'].message_type = _LOCATION
_TRAFFICFLOW.fields_by_name['speed'].message_type = _SPEED
_TRAFFICFLOWWITHPREDICTION.fields_by_name['location'].message_type = _LOCATION
_TRAFFICFLOWWITHPREDICTION.fields_by_name['speed'].message_type = _SPEEDWITHTIMESTAMP
_TRAFFICFLOWWITHPREDICTIONPERSECTION.fields_by_name['location'].message_type = _LOCATION
_TRAFFICFLOWWITHPREDICTIONPERSECTION.fields_by_name['speedVector'].message_type = _SPEEDVECTOR
_SPEED.fields_by_name['trafficCondition'].enum_type = _TRAFFICCONDITION
_SPEEDWITHTIMESTAMP.fields_by_name['speed'].message_type = _SPEED
_SPEEDVECTOR.fields_by_name['sectionSpeed'].message_type = _SECTIONSPEED
_SECTIONSPEED.fields_by_name['speed'].message_type = _SPEED
DESCRIPTOR.message_types_by_name['TrafficFlowGroup'] = _TRAFFICFLOWGROUP
DESCRIPTOR.message_types_by_name['MetaInformation'] = _METAINFORMATION
DESCRIPTOR.message_types_by_name['SupplierAndClientInfo'] = _SUPPLIERANDCLIENTINFO
DESCRIPTOR.message_types_by_name['TrafficFlow'] = _TRAFFICFLOW
DESCRIPTOR.message_types_by_name['TrafficFlowWithPrediction'] = _TRAFFICFLOWWITHPREDICTION
DESCRIPTOR.message_types_by_name['TrafficFlowWithPredictionPerSection'] = _TRAFFICFLOWWITHPREDICTIONPERSECTION
DESCRIPTOR.message_types_by_name['Location'] = _LOCATION
DESCRIPTOR.message_types_by_name['Speed'] = _SPEED
DESCRIPTOR.message_types_by_name['SpeedWithTimeStamp'] = _SPEEDWITHTIMESTAMP
DESCRIPTOR.message_types_by_name['SpeedVector'] = _SPEEDVECTOR
DESCRIPTOR.message_types_by_name['SectionSpeed'] = _SECTIONSPEED
DESCRIPTOR.enum_types_by_name['TrafficCondition'] = _TRAFFICCONDITION

TrafficFlowGroup = _reflection.GeneratedProtocolMessageType('TrafficFlowGroup', (_message.Message,), dict(
  DESCRIPTOR = _TRAFFICFLOWGROUP,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.TrafficFlowGroup)
  ))
_sym_db.RegisterMessage(TrafficFlowGroup)

MetaInformation = _reflection.GeneratedProtocolMessageType('MetaInformation', (_message.Message,), dict(
  DESCRIPTOR = _METAINFORMATION,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.MetaInformation)
  ))
_sym_db.RegisterMessage(MetaInformation)

SupplierAndClientInfo = _reflection.GeneratedProtocolMessageType('SupplierAndClientInfo', (_message.Message,), dict(
  DESCRIPTOR = _SUPPLIERANDCLIENTINFO,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.SupplierAndClientInfo)
  ))
_sym_db.RegisterMessage(SupplierAndClientInfo)

TrafficFlow = _reflection.GeneratedProtocolMessageType('TrafficFlow', (_message.Message,), dict(
  DESCRIPTOR = _TRAFFICFLOW,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.TrafficFlow)
  ))
_sym_db.RegisterMessage(TrafficFlow)

TrafficFlowWithPrediction = _reflection.GeneratedProtocolMessageType('TrafficFlowWithPrediction', (_message.Message,), dict(
  DESCRIPTOR = _TRAFFICFLOWWITHPREDICTION,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.TrafficFlowWithPrediction)
  ))
_sym_db.RegisterMessage(TrafficFlowWithPrediction)

TrafficFlowWithPredictionPerSection = _reflection.GeneratedProtocolMessageType('TrafficFlowWithPredictionPerSection', (_message.Message,), dict(
  DESCRIPTOR = _TRAFFICFLOWWITHPREDICTIONPERSECTION,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.TrafficFlowWithPredictionPerSection)
  ))
_sym_db.RegisterMessage(TrafficFlowWithPredictionPerSection)

Location = _reflection.GeneratedProtocolMessageType('Location', (_message.Message,), dict(
  DESCRIPTOR = _LOCATION,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.Location)
  ))
_sym_db.RegisterMessage(Location)

Speed = _reflection.GeneratedProtocolMessageType('Speed', (_message.Message,), dict(
  DESCRIPTOR = _SPEED,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.Speed)
  ))
_sym_db.RegisterMessage(Speed)

SpeedWithTimeStamp = _reflection.GeneratedProtocolMessageType('SpeedWithTimeStamp', (_message.Message,), dict(
  DESCRIPTOR = _SPEEDWITHTIMESTAMP,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.SpeedWithTimeStamp)
  ))
_sym_db.RegisterMessage(SpeedWithTimeStamp)

SpeedVector = _reflection.GeneratedProtocolMessageType('SpeedVector', (_message.Message,), dict(
  DESCRIPTOR = _SPEEDVECTOR,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.SpeedVector)
  ))
_sym_db.RegisterMessage(SpeedVector)

SectionSpeed = _reflection.GeneratedProtocolMessageType('SectionSpeed', (_message.Message,), dict(
  DESCRIPTOR = _SECTIONSPEED,
  __module__ = 'ProtobufTrafficFlow_v5_pb2'
  # @@protoc_insertion_point(class_scope:trafficFlow.SectionSpeed)
  ))
_sym_db.RegisterMessage(SectionSpeed)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
''' OpenLR decoder and database interfaces

    .. moduleauthor:: David Marteau <david.marteau@mappy.com>

    Pylr use an abstract mapdatabase interface for accessing grapth data.
    The map database is passed to the decoder for calculating route and shortest path between 
    location reference points.
'''
from __future__ import print_function

from collections import namedtuple
from itertools import groupby, chain
from gv_vendor.pylr import rating as Rating
from gv_vendor.pylr.constants import (LocationType,
                                      WITH_LINE_DIRECTION,
                                      AGAINST_LINE_DIRECTION,
                                      BINARY_VERSION_2,
                                      BINARY_VERSION_3)


''' The Max_ node_ distance '''
MAX_NODE_DIST = 100

''''The node factor.'''
NODE_FACTOR = 3

'''The line factor'''
LINE_FACTOR = 3

'''The FRC variance'''
FRC_VARIANCE = 2

'''The minimum accepted rating'''
MIN_ACC_RATING = 800

'''Max number of retries'''
MAX_NR_RETRIES = 3

'''The same line degradation'''
SAME_LINE_DEGRAD = 0.10

'''Connected route increase'''
CONNECT_ROUTE_INC = 0.10

'''DNP variance'''
DNP_VARIANCE = 118

'''FRC rating'''
FRC_RATING = {Rating.EXCELLENT: 100,
              Rating.GOOD: 75,
              Rating.AVERAGE: 50,
              Rating.POOR: 0}

'''FRC_Intervals'''
FRC_INTERVALS = {Rating.EXCELLENT: 0,
                 Rating.GOOD: 1,
                 Rating.AVERAGE: 2}

'''FOW rating'''
FOW_RATING = {Rating.EXCELLENT: 100,
              Rating.GOOD: 50,
              Rating.AVERAGE: 50,
              Rating.POOR: 25}

'''Bearing rating'''
BEAR_RATING = {Rating.EXCELLENT: 100,
               Rating.GOOD: 50,
               Rating.AVERAGE: 25,
               Rating.POOR: 0}

# Compare bearing using integer bearing values

'''Bearing intervals'''
BEAR_INTERVALS = {Rating.EXCELLENT: 0,
                  Rating.GOOD: 1,
                  Rating.AVERAGE: 2}

'''Max bearing diff'''
MAX_BEAR_DIFF = 8  # 90deg

HALF_CIRCLE = 16  # 180deg

FULL_CIRCLE = 32  # 360deg

'''Calc affected lines'''
CALC_AFFECTED_LINES = False

'''Lines directly factor'''
LINES_DIRECTLY_FACTOR = 0.95

'''CompTime4Cache'''
COMP_TIME_4_CACHE = 0

# ----------------
# Special values
# ----------------

# Define empty route
EMPTY_ROUTE = ((), 0)

# ----------------
# Decoder exceptions
# ----------------


class DecoderError(Exception):
    pass


class DecoderInvalidLocation(DecoderError):
    pass


class DecoderNodesNotFound(DecoderError):
    pass


class DecoderNoCandidateLines(DecoderError):
    pass


class RouteSearchException(DecoderError):
    pass


class RouteNotFoundException(RouteSearchException):
    pass


class InvalidRouteLength(RouteNotFoundException):
    pass


class RouteConstructionFailed(RouteSearchException):
    pass

# ----------------
# Map database
# ----------------


class MapDatabase(object):
    """ Abstract interface used by the decoder object.
    
        Implementor of database should inherit from this abstract class.
        
        MapDatabase defines two data structure as named tuples:
        
        :py:class:`MapDatabase.Node`

        :py:class:`MapDatabase.Line`

        These structures may be extended by MapDatabase implementor accordings to their specific needs.          
    """

    Node = namedtuple('Node', ('distance',))
    """
        .. attribute:: distance
       
           The distance from the search location
    """
 
    Line = namedtuple('Line', ('id', 'bear', 'frc', 'fow', 'len', 'projected_len'))
    """
        .. attribute:: id
        
            id of the line
            
        .. attribute:: bear
        
            the bearing according to the start node
            
        .. attribute:: frc
        
             the frc of the line
             
        .. attribute:: fow
        
            the fow of the line

        .. attribute:: projected_len

            return the value of the projected length of the search location
            (i.e) the distance between the start node and the projection of the
            point given by the search coordinates. None if the line is not
            projected
    """

    def connected_lines(self, node, frc_max, beardir):
        """ Return connected lines to/from the node 'node'
        
            :param frc_max: the frc max of the requested lines
            :param beardir: select the inwards (AGAINST_LINE_DIRECTION)
            or the outwards (WITH_LINE_DIRECTION) connected lines

            return an iterable of objects of type Line
        """
        raise NotImplementedError("MapDatabase:connectedlines")

    def find_closeby_nodes(self, coords, max_node_dist):
        """ Look for all nodes at less than max_node_dist from
            the given coordinates
            
            :param coords: an tuple or iterable holding location coordinates
            :param max_node_dist: max distance to nearch for nodes

            return an iterable of Node objects
        """
        raise NotImplementedError("MapDatabase:find_closeby_nodes")

    def find_closeby_lines(self, coords, max_node_dist, frc_max, beardir):
        """ Look for all lines at less than max_node_dist from
            the given coordinates
            
            :param coords: an tuple or iterable holding location coordinates
            :param max_node_dist: max distance to nearch for nodes
            :param frc_max: the frc max of the requested line
            :param beardir: select the inwards (AGAINST_LINE_DIRECTION)
            or the outwards (WITH_LINE_DIRECTION) connected lines

            return an iterable of Line objects
        """
        raise NotImplementedError("MapDatabase:find_closeby_lines")

    def calculate_route(self, l1, l2, maxdist, lfrc, islastrp):
        """ Calculate the shortest paths between two lines
        
            :param l1: the first candidate line to begin the search from
            :param l2: the second candidate line to stop the search to
            :param maxdist: The maximum distance allowed
            :param lfrc: The least frc allowed
            :param islastrp: True if we are calculating the route to the last
            reference point
            :return: (route, length) where route is an iterable holding the lines found
            and length the calculated length  of the route
    
            The method must throw a RouteNotFoundException or a RouteConstructionFailed
            exception in case a route cannot be calculated
        """
        raise NotImplementedError("MapDatabase:calculate_route")

# ----------------
# Decoder
# ----------------


class DecoderBase(object):
    pass


class RatingCalculator(object):
    """ Implement default rating calculation
    """

    RatingDetails = namedtuple('RatingDetails', ('bear_rating', 'frc_rating', 'fow_rating'))

    def _frc_rating(self, frc, linefrc):
        diff = abs(frc - linefrc)
        for cat in (Rating.EXCELLENT, Rating.GOOD, Rating.AVERAGE):
            if diff <= FRC_INTERVALS[cat]:
                return FRC_RATING[cat]
        return FRC_RATING[Rating.POOR]

    def _fow_rating(self, fow, linefow):
        return FOW_RATING[Rating.get_fow_rating_category(fow, linefow)]

    def _distance_rating(self, dist):
        return max(0, self._max_node_dist - round(dist))

    def _bear_rating(self, bearing, linebear):
        diff = abs(bearing - linebear)
        if diff > HALF_CIRCLE:
            diff = FULL_CIRCLE - diff
        if diff > MAX_BEAR_DIFF:
            return -1
        for cat in (Rating.EXCELLENT, Rating.GOOD, Rating.AVERAGE):
            if diff <= BEAR_INTERVALS[cat]:
                return BEAR_RATING[cat]
        return BEAR_RATING[Rating.POOR]

    def rating(self, lrp, line, dist):
        node_rating = self._distance_rating(dist)
        bear_rating = self._bear_rating(lrp.bear, line.bear)
        if bear_rating < 0:
            return -1

        line_rating = self._frc_rating(lrp.frc, line.frc) +\
            self._fow_rating(lrp.fow, line.fow) +\
            bear_rating

        return node_rating*NODE_FACTOR + line_rating*LINE_FACTOR

    def rating_details(self, lrp, line):
        details = self.RatingDetails(bear_rating=self._bear_rating(lrp.bear, line.bear),
                                     frc_rating=self._frc_rating(lrp.frc, line.frc),
                                     fow_rating=self._fow_rating(lrp.fow, line.fow))
        return details


def calculate_pairs(lines1, lines2, lastline, islastrp, islinelocation):
    """ Each LRP might have several candidate lines. In order to find the best
        pair to start with each pair needs to be investigated and rated. The
        rating process includes:
        
            - score of the first candidate line
            - score of the second candidate line
            - connection to the previously calculated path
            - candidate lines shall not be equal
    """
    # check connection with previously calculated path
    for l1, score1 in lines1:
        if lastline is not None and l1.id == lastline.id:
            score1 += CONNECT_ROUTE_INC * score1
        for l2, score2 in lines2:
            if not islastrp and islinelocation and l2.id == l1.id:
                score2 -= SAME_LINE_DEGRAD * score2
            yield (l1, l2), score1*score2


# Check for single line coverage
def singleline(candidates):
    bests = (lines[0] for lrp, lines in candidates)
    sl, _ = bests.next()
    for l, _ in bests:
        if l.id != sl.id:
            return None
    return sl


class ClassicDecoder(DecoderBase, RatingCalculator):
    """ OpenLR location decoder that use an abstract  map object

        See :py:class:`MapDatabase` for map database  interface.

    """
    def __init__(self, map_database,
                 max_node_distance=MAX_NODE_DIST,
                 frc_variance=FRC_VARIANCE,
                 dnp_variance=DNP_VARIANCE,
                 minimum_acc_rating=MIN_ACC_RATING,
                 find_lines_directly=True,
                 max_retry=MAX_NR_RETRIES,
                 verbose=False,
                 logger=lambda m: print(m)):
        """ Initialize the  decoder

            :param map_database: a map database instance
            :param max_node_distance: the maximun distance for candidate nodes
            :param frc_variance: allowed frc variances
            :param minimum_acc_rating: minimum acceptance rating for candidate lines
            :param find_lines_directly: enable direct search of candidate lines
                                    from lrp projection
            :param max_retry: maximum number of retry when searching for route
                between consecutive lines
        """
        self._mdb = map_database
        self._max_node_dist = max_node_distance
        self._frc_var = frc_variance
        self._min_acc_rating = minimum_acc_rating
        self._max_retry = max_retry
        self._dnp_variance = dnp_variance
        self.verbose = verbose
        self.find_lines_directly = find_lines_directly
        self.logger = logger

    @property
    def database(self):
        return self._mdb

    def find_candidate_nodes(self, lrp):
        """ Find candidate nodes for one location reference point.
            The max_node_distance configure the search for nodes being
            a possibility.
        """
        return self._mdb.find_closeby_nodes(lrp.coords, self._max_node_dist)

    def find_candidate_lines(self, lrp, beardir=WITH_LINE_DIRECTION, with_details=False):
        """ Find candidate lines for each location reference point. The candidate
            lines will be rated indicating how good they match the LRP attributes.
            The method will be configured by OpenLR properties.

            The connectedlines method takes a 'beardir' argument indicating
            inwards (AGAINST_LINE_DIRECTION) or outwards (WITH_LINE_DIRECTION) arcs
        """
        frc_max = lrp.frc + self._frc_var
        nodes = list(self.find_candidate_nodes(lrp))

        rating_f = self.rating
        min_acc = self._min_acc_rating

        rating_key = lambda l, r: r
        group_key = lambda l, r: l.id

        candidates = ((l, rating_f(lrp, l, n.distance)) for n in nodes for l in self._mdb.connected_lines(
            n, frc_max=frc_max, beardir=beardir))
        if self.find_lines_directly:
            candidates = chain(candidates, self.find_candidate_lines_directly(
                lrp, frc_max=frc_max, alreadyfound=bool(nodes), beardir=beardir))
            candidates = (max(vals, key=rating_key) for k, vals in groupby(
                sorted(candidates, key=group_key), key=group_key))
        if not with_details:
            candidates = filter(lambda l, r: r >= min_acc, candidates)
        lines = sorted(candidates, key=rating_key, reverse=True)
        if not with_details and not lines:
            raise DecoderNoCandidateLines("No candidate lines found....")

        if with_details:
            lines = [(l, r, self.rating_details(lrp, l)) for l, r, in lines]

        return lines

    def find_candidate_lines_directly(self, lrp, frc_max, alreadyfound=False, beardir=WITH_LINE_DIRECTION):
        """ Find candidate lines directly if no node or line has been detected so
            far. This method tries to find all lines which are around the LRP
            coordinate. The coordinate will be projected onto the line and the
            distance between that projection point and the coordinate shall be small
            (according to the encoder properties). All lines will be rated and
            proposed as candidate lines for the LRP.

            :param lrp: the location reference point (having no candidate lines so far)
            :param alreadyfound: the already found lines
        """
        rating_f = self.rating

        lines = self._mdb.find_closeby_lines(lrp.coords, self._max_node_dist, frc_max=frc_max, beardir=beardir)
        for line, dist in lines:
            rating = rating_f(lrp, line, dist)
            if alreadyfound:
                rating = round(LINES_DIRECTLY_FACTOR * rating)

            yield line, rating

    def resolve_route(self, location, candidates):
        """ Resolves the shortest-paths between each subsequent pair of location
            reference points. The method orders the candidate line pairs for two
            subsequent LRPs and starts with the best rated pair to calculate a
            shortest-path in between. The method further checks the minimum and
            maximum distance criteria for the calculated shortest-path. If one of the
            criteria is not fulfilled the methods tries the next best candidate line
            pair. If no further pair is available the method fails. For each
            subsequent pair of LRPs the start LRP will hold the calculated route
            after finishing this method.

            :param location: the location
            :param candidates: an iterable holding tuples of (lrp,candidate_lines)
        """

        if not isinstance(candidates, (list, tuple)):
            candidates = tuple(candidates)

        sl = singleline(candidates)
        if sl is not None:
            return (((sl,), sl.len),)

        islinelocation = (location.type == LocationType.LINE_LOCATION)

        lastlrp = location.llrp
        lastline, prevlrp = None, None

        routes = ()

        nr_retry = self._max_retry+1

        # iterate over all LRP pairs
        for i, (lrp, lines) in enumerate(candidates[:-1]):
            lrpnext, nextlines = candidates[i+1]
            islastrp = lrpnext is lastlrp
            pairs = sorted(calculate_pairs(lines, nextlines, lastline,
                           islastrp, islinelocation), key=lambda p, r: r, reverse=True)
            # check candidate pairs
            for (l1, l2), _ in pairs[:nr_retry]:
                if self.verbose:
                    self.logger("openlr: computing route ({},{})".format(l1.id, l2.id))
                # handle same start/end.
                if l1.id == l2.id:
                    if islastrp:
                        route = ((l1,), l1.len)
                    else:
                        # Skip this
                        route = EMPTY_ROUTE
                    break  # search finished
                try:
                    # calculate route between start and end and a maximum distance
                    route = self._calculate_route(l1, l2, lrp, islastrp)
                    # Handle change in start index
                    if lastline is not None and lastline.id != l1.id:
                        self._handle_start_change(routes, l1, lrp, prevlrp)
                    break  # search finished
                except (RouteNotFoundException, RouteConstructionFailed):
                    # Let a chance to retry
                    route = None

            if route is None:
                raise RouteNotFoundException("Route not found")

            if route is not EMPTY_ROUTE:
                routes += (route,)

            if self.verbose:
                # Display route
                lines, length = route
                self.logger("openlr: resolved route ({},{}):{} length={}".format(
                    l1.id, l2.id, tuple(l.id for l in lines), length))

            prevlrp, lastline = lrp, l2

        return routes

    def _handle_start_change(self, routes, lend, lrp, prevlrp):
        """ Recompute previous route using new end line
        """
        lstart, _ = routes[-1][0]
        if self.verbose:
            self.logger("openlr: recomputing last route between {} and {}".format(lstart.id, lend.id))
        route = self._calculate_route(lstart, lend, prevlrp, islastrp=False)
        routes = routes[:-1] + (route, )

    def _calculate_route(self, l1, l2, lrp, islastrp):
            """ Calculate shortest-path between two lines
            """
            # determine the minimum frc for the path to be calculated
            lfrc = lrp.lfrcnp + self._frc_var
            # Calculates the maximum allowed distance between two location reference
            # points taking into account that at least one LRP might be projected onto
            # a line and the maximum distance must be adjusted as the route calculation
            # can only stop at distances between real nodes.
            maxdist = lrp.dnp + self._dnp_variance
            # check if LRPs were projected on line (i.e obtained directly)
            # if yes, add line length to maxDistance (complete length as route
            # search stops at nodes)
            if l1.projected_len is not None:
                maxdist += l1.len
            if l2.projected_len is not None:
                maxdist += l2.len
            # calculate route between start and end and a maximum distance
            route, length = self._mdb.calculate_route(l1, l2, maxdist, lfrc, islastrp)
            # adjust and check the route length
            if l2.projected_len is not None:
                if islastrp:
                    length -= l2.len
                length += l2.projected_len
            # check the minimum distance criteria
            if max(0, lrp.dnp - self._dnp_variance) > length:
                raise InvalidRouteLength("openlr: route: {} to {}, calculated length:{}, lrp:{}".format(
                    l1.id, l2.id, length, lrp))

            return route, length

    def calculate_offsets(self, location, routes):
        # Compute offsets
        if location.version == BINARY_VERSION_2:
            return location.poffs, location.noffs
        elif location.version == BINARY_VERSION_3:
            (head, head_len), (tail, tail_len) = routes[0], routes[-1]

            head_start_line = head[0]
            cutstart = 0
            if head_start_line.projected_len is not None:
                cutstart = head_start_line.projected_len
                head_len -= cutstart

            tail_end_line = tail[-1]
            cutend = 0
            if tail_end_line.projected_len is not None:
                cutend = tail_end_line.len - tail_end_line.projected_len
                tail_len -= cutend

            if len(routes) == 1:
                head_len -= cutend
                tail_len -= cutstart
            else:
                # get the first line of the next sub-path
                head_end_line = (routes[1][0])[0]
                if head_end_line.projected_len is not None:
                    # there is another part on the first line of the next
                    # sub-path that relates to the head part
                    head_len += head_end_line.projected_len
                # get the first line of the last sub-path
                tail_start_line = tail[0]
                if tail_start_line.projected_len is not None:
                    # not everything of the tail belongs to it, there is a snippet
                    # at the start that refers to the former sub-path
                    tail_len -= tail_start_line.projected_len

            return (round(location.poffs * head_len / 100.0),
                    round(location.noffs * tail_len / 100.0))

    @staticmethod
    def _prune(pruned, off, index):
        prunedlen = 0
        while len(pruned) > 1:
            line = pruned[index]
            length = line.len
            if prunedlen + length > off:
                break
            else:
                prunedlen += length
                pruned.pop(index)
        return off - prunedlen

    @staticmethod
    def _calculated_path(pruned, poff, noff=0):
        length = sum(l.len for l in pruned)
        return [l.id for l in pruned], length, poff, noff

    def decode_line(self, location):
        """ Decode a line from a list of a location reference points

            return (edges, length, poffset, noffset)
        """
        # assert location.type == LocationType.LINE_LOCATION

        def candidates():
            yield location.flrp, self.find_candidate_lines(location.flrp)
            for lrp in location.points:
                yield lrp, self.find_candidate_lines(lrp)
            yield location.llrp, self.find_candidate_lines(location.llrp, AGAINST_LINE_DIRECTION)

        routes = self.resolve_route(location, candidates())
        poff, noff = self.calculate_offsets(location, routes)

        route_length = sum(length for _, length in routes)

        offsum = noff + poff
        # check for too long offset values
        if offsum >= 2*route_length:
            raise DecoderInvalidLocation("Invalid offsets")

        # prune path
        # The positive offset will be used to shorten the
        # location from the beginning and the negative offset will be used to
        # shorten the location from the end. <br>
        # The pruning will always stop at nodes and there will be no pruning of
        # parts of lines. The remaining offsets can be accessed from the returned
        # decoded location object. Remaining offsets which are below the length
        # variance parameter will be refused and set to 0

        if offsum > route_length:
            # offsets exceed location length
            # decrease offset values
            ratio = route_length / float(offsum)
            poff = round(poff * ratio)
            noff = round(noff * ratio)
            # retainable length shall be 1 meter
            if poff > noff:
                poff -= 1
            else:
                noff -= 1

        pruned = list(chain(*(lines for lines, _ in routes)))

        if poff > 0:
            poff = self._prune(pruned, poff, 0)
        if noff > 0:
            noff = self._prune(pruned, noff, -1)

        return self._calculated_path(pruned, poff, noff)

    def decode_point(self, location):
        """ Decode a point location from a couple of lrps
            return (edges, length, poffset)
        """
        # assert location.type in (LocationType.POINT_LOCATION_TYPES, LocationType.POI_WITH_ACCESS_POINT)

        routes = self.resolve_route(location, ((location.flrp, self.find_candidate_lines(location.flrp)),
                                   (location.llrp, self.find_candidate_lines(location.llrp, AGAINST_LINE_DIRECTION))))

        head, head_len = routes[0]
        lstart, lend = head[0], head[-1]

        prunedlen = 0
        poff = 0

        if lstart.projected_len is not None:
            poff = lstart.projected_len
            prunedlen = poff
        if lend.projected_len is not None:
            prunedlen += lend.len - lend.projected_len

        poff = round(location.poffs*(head_len-prunedlen)/100.0)+poff
        if poff > head_len:
            poff = head_len

        pruned = list(head)
        poff = self._prune(pruned, poff, 0)

        return self._calculated_path(pruned, poff)

    def decode(self, location):
        if location.type == LocationType.LINE_LOCATION:
            return self.decode_line(location)
        else:
            return self.decode_point(location)



# -*- coding: utf-8 -*-
''' Location parser 

    .. moduleauthor:: David Marteau <david.marteau@mappy.com>

'''

from collections import namedtuple
from bitstring import BitStream
from gv_vendor.pylr.utils import lazyproperty
from gv_vendor.pylr.constants import (LATEST_BINARY_VERSION,
                                      BINARY_VERSION_2,
                                      MIN_BYTES_LINE_LOCATION,
                                      MIN_BYTES_POINT_LOCATION,
                                      MIN_BYTES_CLOSED_LINE_LOCATION,
                                      MIN_BYTES_POLYGON,
                                      RELATIVE_COORD_SIZE,
                                      IS_POINT,
                                      HAS_ATTRIBUTES,
                                      GEOCOORD_SIZE,
                                      POINT_ALONG_LINE_SIZE,
                                      POINT_WITH_ACCESS_SIZE,
                                      POINT_OFFSET_SIZE,
                                      AREA_CODE_CIRCLE,
                                      AREA_CODE_RECTANGLE,
                                      AREA_CODE_POLYGON,
                                      RECTANGLE_SIZE,
                                      LARGE_RECTANGLE_SIZE,
                                      GRID_SIZE,
                                      LARGE_GRID_SIZE,
                                      LRP_SIZE,
                                      CIRCLE_BASE_SIZE,
                                      LocationType)


class BinaryParseError(Exception):
    pass


class BinaryVersionError(BinaryParseError):
    pass


class InvalidDataSizeError(BinaryParseError):
    pass


# The Constant RFU (Reserved for Future Use)
RFU_BITS = 'uint:1'

# number of bits used for attributes flag
ATTR_FLAG_BITS = 'uint:1'

# number of bits used for point flag
POINT_FLAG_BITS = 'uint:1'

# number of bits used for version
VERSION_BITS = 'uint:3'

AREA_FLAG_BIT0 = 'uint:1'

AREA_FLAG_BIT1 = 'uint:1'

HEADER_BITS = (RFU_BITS,
               ATTR_FLAG_BITS,
               POINT_FLAG_BITS,
               AREA_FLAG_BIT1,
               POINT_FLAG_BITS,
               VERSION_BITS)


_BinaryHeader = namedtuple('_BinaryHeader', ('arf', 'af', 'pf', 'ver'))


class _RawBinaryData(object):
    """ Hold a location reference description as a bit stream."""

    MIN_VERSION = BINARY_VERSION_2
    MAX_VERSION = LATEST_BINARY_VERSION

    
    def __init__(self, data, base64=False):
        """ Constructor.
        
            :param string data: Binaray data
            :param bool base64: True if data is coded in base64
        """
        if base64:
            data = data.decode("base64")
        
        #: raw data size
        self._sz = len(data)
        
        #: bit stream used to read data
        self._bs = BitStream(bytes=data)

    def getbits(self, *bits):
        """ Read the given numbers of bits.
        
            :param tuple bits: Tuple of number of bits to read
            :returns: Tuple of bit fields
            :rtype: tuple
        """
        return tuple(self._bs.read(v) for v in bits)
    
    def get_position(self):
        """ Returns position in the bit stream.
        
            :returns: Position in the bit stream
            :rtype: int
        """
        return self._bs.pos

    @property
    def num_bytes(self):
        """ Size of the decoded data.
        
            :returns: Size of the decoded data.
            :rtype: int
        """
        return self._sz

    @property
    def version(self):
        """ Return binary version of the data
        
            :returns: Binary version of the data.
            :rtype: int
        """
        return self.header.ver

    @lazyproperty
    def header(self):
        """ Parse header (once) location type
        
            :returns: Header data
            :rtype: _BinaryHeader
        """
        # Validate data size
        if self._sz < min(MIN_BYTES_LINE_LOCATION,
                          MIN_BYTES_POINT_LOCATION,
                          MIN_BYTES_CLOSED_LINE_LOCATION):
            raise InvalidDataSizeError("not enough bytes in data")

        _, arf1, pf, arf0, af, ver = self.getbits(*HEADER_BITS)
        arf = 2 * arf1 + arf0

        return _BinaryHeader(arf, af, pf, ver)

    @lazyproperty
    def location_type(self):
        """ Parse location type (once)
        
            :returns: Location type
            :rtype: LocationType
        """
        header = self.header

        # Check version
        if not self.MIN_VERSION <= header.ver <= self.MAX_VERSION:
            raise BinaryVersionError("Invalid binary version {}".format(header.ver))

        is_point = (header.pf == IS_POINT)
        has_attributes = (header.af == HAS_ATTRIBUTES)
        area_code = header.arf
        is_area = ((area_code == 0 and not is_point and not has_attributes) or area_code > 0)

        total_bytes = self._sz

        loc_type = LocationType.UNKNOWN

        if not is_point and not is_area and has_attributes:
            loc_type = LocationType.LINE_LOCATION
        elif is_point and not is_area:
            if not has_attributes:
                if total_bytes == GEOCOORD_SIZE:
                    loc_type = LocationType.GEO_COORDINATES
                else:
                    raise InvalidDataSizeError("Invalid byte size")
            else:
                if total_bytes == POINT_ALONG_LINE_SIZE or total_bytes == (POINT_ALONG_LINE_SIZE + POINT_OFFSET_SIZE):
                    loc_type = LocationType.POINT_ALONG_LINE
                elif total_bytes == POINT_WITH_ACCESS_SIZE or total_bytes == (POINT_WITH_ACCESS_SIZE + POINT_OFFSET_SIZE):
                    loc_type = LocationType.POI_WITH_ACCESS_POINT
                else:
                    raise InvalidDataSizeError("Invalid byte size")
        elif is_area and not is_point and has_attributes:
            if total_bytes >= MIN_BYTES_CLOSED_LINE_LOCATION:
                loc_type = LocationType.CLOSED_LINE
            else:
                raise InvalidDataSizeError("Invalid byte size")
        else:
            if area_code == AREA_CODE_CIRCLE:
                loc_type = LocationType.CIRCLE
            elif area_code == AREA_CODE_RECTANGLE:
                # includes case AREA_CODE_GRID
                if total_bytes == RECTANGLE_SIZE or total_bytes == LARGE_RECTANGLE_SIZE:
                    loc_type = LocationType.RECTANGLE
                elif total_bytes == GRID_SIZE or total_bytes == LARGE_GRID_SIZE:
                    loc_type = LocationType.GRID
                else:
                    raise InvalidDataSizeError("Invalid byte size")
            elif area_code == AREA_CODE_POLYGON:
                if not has_attributes and total_bytes >= MIN_BYTES_POLYGON:
                    loc_type = LocationType.POLYGON
                else:
                    raise InvalidDataSizeError("Invalid byte size")
            else:
                raise BinaryParseError('Invalid header')

        return loc_type


def init_binary_parsing(data, base64=False):
    """ Create an instance of _RawBinaryData
        The returned object can be passed to 'parse_binary'
        
        :param string data: string describing the location
        :param bool base64: True if encoded in base 64
        :returns: Parsable data structure
        :rtype: _RawBinaryData
    """
    return _RawBinaryData(data, base64)


def parse_binary(data, base64=False):
    """ Parse binary data.
        Input is original data or an object returned by init_binary_parsing(...)
        
        :param data: string (encoded or not) describing the location
        :param bool base64: True if encoded in base 64
        :returns: Object describing the parsed location, or an error object
    """
    if not isinstance(data, _RawBinaryData):
        data = _RawBinaryData(data, base64)

    # Get header
    loc_type = data.location_type
    
    if loc_type == LocationType.LINE_LOCATION:
        return parse_line(data)
    elif loc_type == LocationType.POINT_ALONG_LINE:
        return parse_point_along_line(data)
    elif loc_type == LocationType.GEO_COORDINATES:
        return parse_geo_coordinates(data)
    elif loc_type == LocationType.POI_WITH_ACCESS_POINT:
        return parse_poi_with_access_point(data)
    elif loc_type == LocationType.RECTANGLE:
        return parse_rectangle(data)
    elif loc_type == LocationType.CLOSED_LINE:
        return parse_closed_line(data)
    elif loc_type == LocationType.CIRCLE:
        return parse_circle(data)
    elif loc_type == LocationType.GRID:
        return parse_grid(data)
    elif loc_type == LocationType.POLYGON:
        return parse_polygon(data)
    else:
        return BinaryParseError("Invalid location type")

# ----------------
# Location parsers
# ----------------

HEAD_FIELDS = ('version', 'type')

from gv_vendor.pylr.binary import (_parse_first_lrp,
                                   _parse_intermediate_lrp,
                                   _parse_last_line_lrp,
                                   _parse_last_closed_line_attrs,
                                   _parse_offset,
                                   _parse_relative_coordinates,
                                   _parse_absolute_coordinates,
                                   _parse_radius,
                                   _parse_grid_dimensions)


# LINE_LOCATION
LineLocation = namedtuple('LineLocation',  HEAD_FIELDS+('flrp', 'llrp', 'points', 'poffs', 'noffs'))
""" Line Location type
"""

def parse_line(rb):
    """ Parse line location
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Line location
        :rtype: LineLocation
    """
    assert rb.location_type == LocationType.LINE_LOCATION

    # number of intermediates points
    num_intermediates = int((rb.num_bytes - MIN_BYTES_LINE_LOCATION) / LRP_SIZE)
    flrp = _parse_first_lrp(rb)

    points = []

    rel = flrp
    for _ in range(num_intermediates):
        ilrp = _parse_intermediate_lrp(rb, rel)
        points.append(ilrp)
        rel = ilrp

    llrp, pofff, nofff = _parse_last_line_lrp(rb, rel)
    poffs = _parse_offset(rb) if pofff else 0
    noffs = _parse_offset(rb) if nofff else 0

    return LineLocation(rb.version, rb.location_type, flrp, llrp, points, poffs, noffs)


# POINT_ALONG_LINE
PointAlongLineLocation = namedtuple('PointAlongLineLocation',  HEAD_FIELDS+('flrp', 'llrp', 'poffs'))
""" Point along location type
"""

def parse_point_along_line(rb):
    """ Parse point along line location
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Point along line location
        :rtype: PointAlongLineLocation
    """
    assert rb.location_type == LocationType.POINT_ALONG_LINE

    flrp = _parse_first_lrp(rb)
    llrp, pofff, _ = _parse_last_line_lrp(rb, flrp)
    poffs = _parse_offset(rb) if pofff else 0

    return PointAlongLineLocation(rb.version, rb.location_type, flrp, llrp, poffs)


# GEO_COORDINATES
GeoCoordinateLocation = namedtuple('GeoCoordinateLocation',  HEAD_FIELDS+('coords',))
""" Coordinate location type
"""

def parse_geo_coordinates(rb):
    """ Parse geo coordinates location
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Geographic coordinates location
        :rtype: GeoCoordinateLocation
    """
    assert rb.location_type == LocationType.GEO_COORDINATES

    coords = _parse_absolute_coordinates(rb)

    return GeoCoordinateLocation(rb.version, rb.location_type, coords)


# POI_WITH_ACCESS_POINT
PoiWithAccessPointLocation = namedtuple('PoiWithAccessPointLocation',  HEAD_FIELDS+(
    'flrp', 'llrp', 'poffs', 'coords'))
""" Poi with access location type
"""

def parse_poi_with_access_point(rb):
    """ Parse POI with access point
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: POI with access point location
        :rtype: PoiWithAccessPointLocation
    """
    assert rb.location_type == LocationType.POI_WITH_ACCESS_POINT

    flrp = _parse_first_lrp(rb)
    llrp, pofff, _ = _parse_last_line_lrp(rb, flrp)
    poffs = _parse_offset(rb) if pofff else 0
    coords = _parse_relative_coordinates(rb, flrp.coords)

    return PoiWithAccessPointLocation(rb.version, rb.location_type, flrp, llrp,
                                      poffs, coords)


# CIRCLE
CircleLocation = namedtuple('CircleLocation', HEAD_FIELDS+('coords', 'radius'))
""" Circle Location type
"""


def parse_circle(rb):
    """ Parse circle location
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Circle location
        :rtype: CircleLocation
    """
    assert rb.location_type == LocationType.CIRCLE

    radius_size = rb.num_bytes - CIRCLE_BASE_SIZE
    
    coords = _parse_absolute_coordinates(rb)
    radius = _parse_radius(rb, radius_size)
    
    return CircleLocation(rb.version, rb.location_type, coords, radius)


# RECTANGLE

BBox = namedtuple('BBox', ('minx', 'miny', 'maxx', 'maxy'))

RectangleLocation = namedtuple('RectangleLocation',  HEAD_FIELDS+('bbox',))
""" Rectangle Location type
"""

def parse_rectangle(rb):
    """ Parse rectangle location
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Rectangle location
        :rtype: RectangleLocation
    """
    assert rb.location_type == LocationType.RECTANGLE

    bl = _parse_absolute_coordinates(rb)
    if rb.num_bytes == LARGE_RECTANGLE_SIZE:
        tr = _parse_absolute_coordinates(rb)
    else:
        tr = _parse_relative_coordinates(rb, bl)

    bbox = BBox(bl.lon, bl.lat, tr.lon, tr.lat)

    return RectangleLocation(rb.version, rb.location_type, bbox)


# GRID
GridLocation = namedtuple('GridLocation',  HEAD_FIELDS+('bbox', 'cols', 'rows'))
""" Grid Location type
"""

def parse_grid(rb):
    """ Parse grid location
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Grid location
        :rtype: GridLocation
    """
    assert rb.location_type == LocationType.GRID

    bl = _parse_absolute_coordinates(rb)
    if rb.num_bytes == LARGE_GRID_SIZE:
        tr = _parse_absolute_coordinates(rb)
    else:
        tr = _parse_relative_coordinates(rb, bl)

    bbox = BBox(bl.lon, bl.lat, tr.lon, tr.lat)

    cols, rows = _parse_grid_dimensions(rb)

    return GridLocation(rb.version, rb.location_type, bbox, cols, rows)


# CLOSED LINE
ClosedLineLocation = namedtuple('ClosedLineLocation',  HEAD_FIELDS+('flrp', 'points', 'frc', 'fow', 'bear'))

def parse_closed_line(rb):
    """ Parse closed line location
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Closed line location
        :rtype: ClosedLineLocation
    """
    assert rb.location_type == LocationType.CLOSED_LINE

    # number of intermediates points
    num_intermediates = (rb.num_bytes - MIN_BYTES_CLOSED_LINE_LOCATION) / LRP_SIZE
    flrp = _parse_first_lrp(rb)
    
    points = []

    rel = flrp
    for _ in range(num_intermediates):
        ilrp = _parse_intermediate_lrp(rb, rel)
        points.append(ilrp)
        rel = ilrp

    frc, fow, bear = _parse_last_closed_line_attrs(rb)
    
    return ClosedLineLocation(rb.version, rb.location_type, flrp, points, frc, fow, bear)


# CLOSED LINE
PolygonLocation = namedtuple('PolygonLocation',  HEAD_FIELDS+('points',))

def parse_polygon(rb):
    """ Parse polygon location
    
        :param _RawBinaryData rb: Binary data describing the location
        :returns: Polygon location
        :rtype: PolygonLocation
    """
    assert rb.location_type == LocationType.POLYGON

    # number of points
    # MIN_BYTES_POLYGON include first point and 2 relatives points
    num_intermediates = 2 + (rb.num_bytes - MIN_BYTES_POLYGON) / RELATIVE_COORD_SIZE
    points = []

    rel = _parse_absolute_coordinates(rb)
    points.append(rel)
    for _ in range(num_intermediates):
        ilrp = _parse_relative_coordinates(rb, rel)
        points.append(ilrp)
        rel = ilrp

    return PolygonLocation(rb.version, rb.location_type, points)



#!/usr/bin/env python3

import asyncio
import traceback
from typing import Awaitable, Callable

import aioredis

from gv_utils import parser
from gv_utils.enums import Message
from gv_utils.logger import Logger


class PubSub:

    def __init__(self, logger: Logger, custom_data_decoder: Callable = None) -> None:
        self.logger = logger
        self.redis = None
        self.custom_data_decoder = custom_data_decoder

    async def async_init(self, redisaddr: str) -> None:
        self.redis = await aioredis.create_redis_pool(redisaddr)

    async def publish(self, data: object, dataformat: str, datatimestamp, datakind: str) -> bool:
        success = False
        try:
            message = {
                Message.data: data,
                Message.format: dataformat,
                Message.timestamp: datatimestamp,
                Message.kind: datakind
            }
            await self.redis.publish(parser.add_main_prefix(datakind), await parser.encode_message(message))
            success = True
        except:
            self.logger.error(traceback.format_exc())
            self.logger.error('An error occurred while publishing {}.'.format(datakind))
        finally:
            return success

    async def subscribe(self, datakind: str, callback: Callable[[dict], Awaitable]) -> None:
        channel, = await self.redis.psubscribe(datakind)
        self.logger.info('RPC client has subscribed to {}.'.format(datakind))
        while await channel.wait_message():
            self.logger.debug('New {}.'.format(datakind))
            try:
                _, message = await channel.get(encoding='utf-8')
                asyncio.create_task(callback(await parser.decode_message(message, self.custom_data_decoder)))
            except:
                self.logger.error(traceback.format_exc())
                self.logger.error('An error occurred while handling {}.'.format(datakind))
        self.logger.info('RPC client has unsubscribed from {}.'.format(datakind))

    async def close(self) -> None:
        if self.redis is not None:
            self.redis.close()
            await self.redis.wait_closed()



#!/usr/bin/env python3

import os

import aiofiles

from gv_utils import datetime
from gv_utils.enums import MessageKind

DATA_PATH_STRUCT = '%Y/%m/%d/'


async def write_data(basepath: str, datafilestruct: str, data: bytes, datakind: str, datadate: datetime) -> str:
    fullpath = os.path.join(_get_path(basepath, datakind, datadate), _get_file_name(datadate, datafilestruct))
    await _write_bytes(fullpath, data)
    return fullpath


async def read_data(basepath: str, datafilestruct: str, datakind: str, datadate: datetime) -> bytes:
    return await _read_bytes(_get_file_path(basepath, datafilestruct, datakind, datadate))


async def data_exists(basepath: str, datafilestruct: str, datakind: str, datadate: datetime) -> bool:
    return os.path.exists(_get_file_path(basepath, datafilestruct, datakind, datadate))


def _get_path(basepath: str, datakind: str, datadate: datetime) -> str:
    return mkdir_if_not_exist(os.path.join(basepath, _get_datakind_dir(datakind),
                                           datetime.to_str(datadate, DATA_PATH_STRUCT)))


def _get_file_name(datadate: datetime, datafilestruct: str) -> str:
    return datetime.to_str(datadate, datafilestruct)


def _get_datakind_dir(datakind: str) -> str:
    sep = MessageKind.separator
    splited = datakind.split(sep)
    datakind = sep.join(splited[-1:] + splited[:-1])
    return datakind.upper()


def _get_file_path(basepath: str, datafilestruct: str, datakind: str, datadate: datetime) -> str:
    return os.path.join(basepath, _get_datakind_dir(datakind),
                        datetime.to_str(datadate, DATA_PATH_STRUCT + datafilestruct))


async def _write_bytes(path: str, b: bytes) -> None:
    async with aiofiles.open(path, 'wb') as file:
        await file.write(b)


async def _read_bytes(path: str) -> bytes:
    if os.path.exists(path):
        async with aiofiles.open(path, 'rb') as file:
            b = await file.read()
    else:
        b = b''
    return b


def join_if_not_abs(root: str, path: str) -> str:
    if not os.path.isabs(path):
        path = os.path.join(root, path)
    return path


def mkdir_if_not_exist(path: str) -> str:
    if not os.path.exists(path):
        os.makedirs(path)
    return path



#!/usr/bin/env python3

import os

from dotenv import load_dotenv


def load_env_var(filepath: str) -> str:
    projectroot = os.path.dirname(os.path.dirname(os.path.realpath(filepath)))

    dotenv_path = os.path.join(projectroot, '.env')
    load_dotenv(dotenv_path)
    envtype = os.environ.get('ENV_TYPE', default='local')
    load_dotenv(dotenv_path + '.' + envtype)

    return projectroot



# -*- coding: utf-8 -*-
''' Some utilities

    .. moduleauthor:: David Marteau <david.marteau@mappy.com>
    
'''


def enum(*sequential, **named):
    """ Create an enum type, as in the C language.
        You can, as i C, list names without any value, or set a value by using a named argument.
        But, of course, you cannot add unnamed arguments after named arguments...
        
        :param sequential: list of strings
        :param named: dictionnary of <strings:int value>
        :return: A new type created by using the params.
        :rtype: Enum  
    """
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)


def signum(x):
    """ Give info about the sign of the given int.
    
        :param int x: Value whose sign is asked
        :return: -1 if the given value is negative, 0 if it is null, 1 if it is positive
        :rtype: int
    """
    return -1 if x < 0 else 0 if x == 0 else 1


class lazyproperty(object):
    """ Meant to be used as decorator for lazy evaluation of an object attribute.
        Property should represent non-mutable data, as it replaces itself.
    """
    def __init__(self, fget):
        self.fget = fget
        self.func_name = fget.__name__
        self.__doc__ = fget.__doc__

    def __get__(self, obj, cls):
        if obj is None:
            return None
        value = self.fget(obj)
        setattr(obj, self.func_name, value)
        return value



# -*- coding: utf-8 -*-
'''
Created on 4 déc. 2013

.. moduleauthor:: David Marteau <david.marteau@mappy.com>

OpenLR decoder

'''

from gv_vendor.pylr.version import __version_info__, __version_info__

from gv_vendor.pylr.parser import (BinaryParseError,
                                   BinaryVersionError,
                                   InvalidDataSizeError,
                                   init_binary_parsing,
                                   parse_binary,
                                   LocationType,
                                   LineLocation,
                                   GeoCoordinateLocation,
                                   BBox,
                                   CircleLocation,
                                   RectangleLocation,
                                   GridLocation,
                                   PolygonLocation,
                                   PoiWithAccessPointLocation,
                                   ClosedLineLocation,
                                   PointAlongLineLocation,
                                   GeoCoordinateLocation,
                                   PoiWithAccessPointLocation,
                                   RectangleLocation,
                                   BBox)

from gv_vendor.pylr.binary import (Coords,
                                   LocationReferencePoint)

from gv_vendor.pylr.constants import (LocationType,
                                      LENGTH_INTERVAL,
                                      BEARING_SECTOR,
                                      RELATIVE_OFFSET_LENGTH,
                                      RIGTH_SIDE,
                                      LEFT_SIDE,
                                      BOTH_SIDE,
                                      NO_ORIENTATION_OR_UNKNOWN,
                                      WITH_LINE_DIRECTION,
                                      AGAINST_LINE_DIRECTION,
                                      BOTH_DIRECTIONS,
                                      ON_ROAD_OR_UNKNOWN)



from gv_vendor.pylr import fow


from gv_vendor.pylr.decoder import (DecoderError,
                                    DecoderInvalidLocation,
                                    RouteSearchException,
                                    RouteNotFoundException,
                                    RouteConstructionFailed,
                                    MapDatabase,
                                    DecoderBase,
                                    RatingCalculator,
                                    ClassicDecoder)

Decoder = ClassicDecoder
